#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <gmp.h>

/* Uses some gcc extensions, so gcc is required.  Also link with libgmp.
 * I used the following command line:

gcc -O9 -march=native -fomit-frame-pointer -finline-functions-called-once -funsafe-loop-optimizations -freorder-blocks-and-partition -fwhole-program -ffast-math -fno-stack-protector -lgmp proper-test-simple.c -o proper-test-simple

*/

/*
 * Uncomment the right include file to test that particular degree
 */
//#include "proper-test-d9.h"
//#include "proper-test-d11.h"
#include "proper-test-d13.h"
//#include "proper-test-d15.h"
//#include "proper-test-d17.h"

/* To only search a subset of the search space,
 * we will search all combinations BEGINOFF,*
 * through ENDOFF,*.  if ENDOFF is undefined, no
 * end condition is checked */
#define BEGINOFF 1
/* #define ENDOFF 9999 */

/* If we want to start with a specific combination to check
 * rather than use BEGINOFF.  Note that these are combinations
 * after removing pure and unreachable terms.
 * Use initc[]={0} to use BEGINOFF. */
/*int initc[] = {6,8,26,59,68,97,103,113}; */
int initc[] = {0};

/* We first do elimination using this modulo.
 * Moduli that work well seem to be 17, 19, 23
 * 13 seems to be too small, large primes mean large
 * lookup table which means cache misses */
#define MODULO 19

#define FALSE 0
#define TRUE (!FALSE)

/* which terms are pure */
int purex[DEGREE-1];
int purey[DEGREE-1];

/* which terms will not appear */
#define UNREACHABLES ((DEGREE-3)/2)
int unreachable[UNREACHABLES];

/* this is without the last column.  It's the number of
 * columns without the pures, unreachables and the last column. */
#define LOGCOLS (COLS-2*(DEGREE-1)-UNREACHABLES-1)

/* translation vector for the original column numbers */
int tr[LOGCOLS];

/* the number of terms the polynomial should have */
#define TARGETN (((DEGREE+3)/2)+((DEGREE+3)%2))

/* the submatrix */
#define ROWS1 (ROWS-2)
#define COLS1 (TARGETN-2+1)

unsigned int modP_submatrix[ROWS1][COLS1];
unsigned int modP_matrix[ROWS][COLS];
__mpz_struct gmp_submatrix[ROWS1][COLS1];

typedef struct {
	int pivr[COLS1];
	int pivc[COLS1];
	int pivots;
} PivotsType;

/* tables for x,y,a,b -> xb-ya mod MODULO? */
static int tblP[MODULO][MODULO][MODULO][MODULO];

static void
precompute_mod_tblP (void)
{
	int i, j, k, l;

	for (i=0; i < MODULO; i++)
		for (j=0; j < MODULO; j++)
			for (k=0; k < MODULO; k++)
				for (l=0; l < MODULO; l++)
					tblP[i][j][k][l] = (i*l+(MODULO-1)*j*k)%MODULO;
}

static void
precompute_mod_matrixP(unsigned int out[ROWS][COLS], int in[ROWS][COLS])
{
	int i, j;

	for (i=0; i < ROWS; i++) {
		for (j=0; j < COLS; j++) {
			int e = in[i][j];
			if (e != 0) {
				while (e < 0) {
					e += MODULO;
				}
			}
			out[i][j] = e % MODULO;
		}
	}
}

static void
swap_rows_mod (unsigned int m[ROWS1][COLS1], int r1, int r2)
{
	unsigned int tmp[COLS1];

	memcpy(tmp, m[r1], sizeof(unsigned int)*COLS1);
	memcpy(m[r1], m[r2], sizeof(unsigned int)*COLS1);
	memcpy(m[r2], tmp, sizeof(unsigned int)*COLS1);
}

static int
modP_ref_is_full_rank (unsigned int m[ROWS1][COLS1])
{
	int i, j, k, d;
	unsigned int x, y;

	d = 0;
	for (j = 0; j < COLS1; j++) {
		for (i = d; i < ROWS1; i++) {
			if (m[i][j] != 0) break;
		}

		if (i == ROWS1) {
			return FALSE;
		} else if (i > d) {
			swap_rows_mod (m, i, d);
		}

		x = m[d][j];

		for (i = d+1; i < ROWS1; i++) {
			y = m[i][j];

			if (y != 0) {
				m[i][j] = 0;
				for (k = j+1; k < COLS1; k++) {
					m[i][k] =
						tblP[x][y][m[d][k]][m[i][k]];
						/* (x*m[i][k] + (MODULO-1)* y*m[d][k]) % MODULO; */
				}
			}
		}

		d++;
	}

	/* found pivots on all columns, full rank */
	return TRUE;
}

static void
swap_rows_gmp (__mpz_struct m[ROWS1][COLS1], int r1, int r2)
{
	__mpz_struct tmp[COLS1];

	memcpy(tmp, m[r1], sizeof(__mpz_struct)*COLS1);
	memcpy(m[r1], m[r2], sizeof(__mpz_struct)*COLS1);
	memcpy(m[r2], tmp, sizeof(__mpz_struct)*COLS1);
}

/* Compute rref but quits before backsubstitution
 * if nullspace more than one dimensional if dim_one_only
 * is TRUE */
static int
gmp_int_rref (__mpz_struct m[ROWS1][COLS1],
	      PivotsType *p,
	      int dim_one_only)
{
	int i, j, k, l, d;
	int mg_set;
	int mg_one;
	mpz_t mg;
	mpz_t tmp;
	mpz_t x;
	mpz_t y;

	p->pivots = 0;

	mpz_init (mg);
	mpz_init (tmp);
	mpz_init (x);
	mpz_init (y);

	d = 0;
	for (j = 0; j < COLS1; j++) {
		for (i = d; i < ROWS1; i++) {
			if (mpz_sgn (&(m[i][j])) != 0) break;
		}

		if (i == ROWS1) {
			continue;
		} else if (i > d) {
			swap_rows_gmp (m, i, d);
		}

		p->pivr[p->pivots] = d;
		p->pivc[p->pivots] = j;
		p->pivots++;

		/* x is the pivot */
		mpz_set (x, &(m[d][j]));

		for (i = d+1; i < ROWS1; i++) {
			if (mpz_sgn (&(m[i][j])) == 0) continue;

			mpz_set (y, &(m[i][j]));

			mg_set = FALSE;
			mg_one = FALSE;

			mpz_set_ui (&(m[i][j]), 0);
			for (k = j+1; k < COLS1; k++) {
				mpz_mul (&(m[i][k]), x, &(m[i][k]));
				mpz_mul (tmp, y, &(m[d][k]));
				mpz_sub (&(m[i][k]), &(m[i][k]), tmp);

				if (mpz_sgn (&(m[i][k])) != 0 &&
				    ! mg_one) {
					if ( ! mg_set) {
						mpz_abs (mg, &(m[i][k]));
						mg_set = TRUE;
					} else {
						mpz_abs (tmp, &(m[i][k]));
						mpz_gcd (mg, mg, tmp);
					}
					if (mpz_cmp_ui (mg, 1) == 0)
						mg_one = TRUE;
				}
			}

			if (mg_set && ! mg_one) {
				for (k = j+1; k < COLS1; k++) {
					mpz_divexact (&(m[i][k]),
						      &(m[i][k]),
						      mg);
				}
			}
		}

		d++;
	}

	if (dim_one_only && (COLS1 - p->pivots) != 1) {
		mpz_clear (mg);
		mpz_clear (tmp);
		mpz_clear (x);
		mpz_clear (y);
		return FALSE;
	}

	for (l = 0; l < p->pivots; l++) {
		d = p->pivr[l];
		j = p->pivc[l];

		mpz_set (x, &(m[d][j]));
		for (i = d-1; i >= 0; i--) {
			if (mpz_sgn (&(m[i][j])) == 0) continue;

			mpz_set (y, &(m[i][j]));

			for (k = 0; k < j; k++) {
				mpz_mul (&(m[i][k]), x, &(m[i][k]));
			}
			mpz_set_ui (&(m[i][j]), 0);
			for (k = j+1; k < COLS1; k++) {
				mpz_mul (&(m[i][k]), x, &(m[i][k]));
				mpz_mul (tmp, y, &(m[d][k]));
				mpz_sub (&(m[i][k]), &(m[i][k]), tmp);
			}
		}
	}

	mpz_clear (mg);
	mpz_clear (tmp);
	mpz_clear (x);
	mpz_clear (y);
	return TRUE;
}

static int
gmp_null_vector (__mpz_struct m[ROWS1][COLS1], __mpz_struct *v, PivotsType *p)
{
	int k, i, jj;
	__mpz_struct ty[COLS1];
	mpz_t my;
	mpz_t a;
	mpz_t b;
	mpz_t g;

	for (i = 0; i < COLS1; i++)
		mpz_init (&(ty[i]));
	mpz_init (my);
	mpz_init (a);
	mpz_init (b);
	mpz_init (g);

	/* loop over nonpivots */
	jj = 0;
	for (k = 0; k < COLS1; k++) {
		/* skip pivots */
		if (jj < p->pivots &&
		    k == p->pivc[jj]) {
			jj++;
			continue;
		}

		for (i = 0; i < COLS1; i++) {
			mpz_set_ui (&(v[i]), 0);
			mpz_set_ui (&(ty[i]), 0);
		}
		
		mpz_set_ui (my, 1);

		for (i = 0; i < p->pivots; i++) {
			int pc = p->pivc[i];
			int pr = p->pivr[i];

			mpz_abs (a, &(m[pr][pc]));
			mpz_abs (b, &(m[pr][k]));
			mpz_gcd (g, a, b);
			mpz_divexact (a, &(m[pr][k]), g);
			mpz_divexact (b, &(m[pr][pc]), g);

			if (mpz_sgn (b) < 0) {
				mpz_neg (b, b);
				mpz_neg (a, a);
			}

			mpz_lcm (my, my, b);
			mpz_neg (&(v[pc]), a);
			mpz_set (&(ty[pc]), b);
		}

		mpz_set (&(v[k]), my);

		for (i = 0; i < p->pivots; i++) {
			int pc = p->pivc[i];
			mpz_mul (&(v[pc]), &(v[pc]), my);
			mpz_div (&(v[pc]), &(v[pc]), &(ty[pc]));
		}

		for (i = 0; i < COLS1; i++)
			mpz_clear (&(ty[i]));
		mpz_clear (my);
		mpz_clear (a);
		mpz_clear (b);
		mpz_clear (g);

		return TRUE;
	}

	for (i = 0; i < COLS1; i++)
		mpz_clear (&(ty[i]));
	mpz_clear (my);
	mpz_clear (a);
	mpz_clear (b);
	mpz_clear (g);

	return FALSE;
}

static int
get_the_modP_submatrix (unsigned int psm[ROWS1][COLS1],
			unsigned int pm[ROWS][COLS],
			int m[ROWS][COLS],
			int c[], int cl)
{
	int i, j;
	
	/* use columns from the column list */
	for (i = 0; i < ROWS-2; i++) {
		int isnull = TRUE;
		for (j = 0; j < cl; j++) {
			int cc = tr[c[j]-1]-1;
			if (m[i+1][cc] != 0) {
				isnull = FALSE;
				psm[i][j] = pm[i+1][cc];
			} else {
				psm[i][j] = 0;
			}
		}
		if (isnull)
			return FALSE;
		/* tack on the last column, that is always positive */
		psm[i][cl] = pm[i+1][COLS-1];
	}
	return TRUE;
}

static void
gmp_get_the_submatrix (__mpz_struct sm[ROWS1][COLS1], int m[ROWS][COLS],
		       int c[], int cl)
{
	int i, j;
	
	/* use columns from the column list */
	for (i = 0; i < ROWS-2; i++) {
		for (j = 0; j < cl; j++) {
			int cc = tr[c[j]-1]-1;
			mpz_set_si (&(sm[i][j]), m[i+1][cc]);
		}
		/* tack on the last column */
		mpz_set_si (&(sm[i][cl]), m[i+1][COLS-1]);
	}
}

static int
gmp_vector_nonnegative (__mpz_struct v[], int len)
{
	int i;
	for (i = 0; i < len; i++) {
		if (mpz_sgn (&(v[i])) < 0)
			return FALSE;
	}
	return TRUE;
}

static int
gmp_vector_nonpositive (__mpz_struct v[], int len)
{
	int i;
	for (i = 0; i < len; i++) {
		if (mpz_sgn (&(v[i])) > 0)
			return FALSE;
	}
	return TRUE;
}

static int
check_pure_unreachable (int cc)
{
	int j;

	for (j=0; j < DEGREE-1; j++) {
		if (purex[j] == cc ||
		    purey[j] == cc)
			return TRUE;
	}
	for (j=0; j < UNREACHABLES; j++) {
		if (unreachable[j] == cc)
			return TRUE;
	}

	return FALSE;
}

static int
checkadjecent (int *c, int cl)
{
	/* Note that since pures have been removed, adjacents are
	 * really adjecent and undoable, so we need not check them */
	int i;

	/* This ought to be smarter, these are all sorted! */
	for (i=0; i < cl-1; i++) {
		int cc1 = c[i];
		int cc2 = c[i+1];
		if (tr[cc1-1]+1 == tr[cc2-1])
			return TRUE;
	}
	return FALSE;
}

static int
comb_get_next_combination (int *vec, int len, int n)
{
	int i = len;
	int j;

	while (i > 0 && vec[i-1] == n-(len-i)) {
		i--;
	}
	if (__builtin_expect (i == 0, 0)) {
		return FALSE;
	} else {
		vec[i-1] ++;
		for (j = i+1; j <= len; j++)
			vec[j-1] = vec[j-2]+1;
	}
	return TRUE;
}

int
main (void)
{
	int i, j;
	PivotsType p;
	int c[TARGETN-2] = {0};
	__mpz_struct v[TARGETN-2+1];
	int cnt = 0;
	int startof_dminusone;
	long long allcnt = 0;
	long long notopcnt = 0;
	long long adjcnt = 0;
	long long nrowcnt = 0;
	long long modfullrankcnt = 0;
	long long finalcnt = 0;
	int begin_time, end_time;

	begin_time = (int)time (NULL);
	printf ("TIME: %d\n", begin_time);

	/* compute and print the purex terms */
	printf ("purex: ");
	j=0;
	for (i=0; i < COLS-1; i++) {
		if (the_matrix[0][i] != 0) {
			purex[j++] = i+1;
			printf ("%d ", i+1);
		}
	}
	startof_dminusone = purex[DEGREE-2];

	printf ("\nSTARTOF D-1 %d\n", startof_dminusone);

	/* compute and print the unreachable terms */
	printf ("un: ");
	j = startof_dminusone+2;
	for (i=0; i < UNREACHABLES; i++) {
		unreachable[i] = j;
		printf ("%d ", j);
		j += 2;
	}
	printf ("\n");

	/* compute and print the purey terms */
	printf ("purey: ");
	j=0;
	for (i=0; i < COLS-1; i++) {
		if (the_matrix[ROWS-1][i] != 0) {
			purey[j++] = i+1;
			printf ("%d ", i+1);
		}
	}

	printf ("\n");

	/* Create the translation table and print it */
	printf ("tr(%d vs %d): ", LOGCOLS, COLS-1);
	j=1;
	for (i=0; i < LOGCOLS; i++) {
		while (check_pure_unreachable (j))
			j++;
		tr[i] = j;
		printf ("%d ", j);
		j++;
	}
	printf ("\n");

	/* use the starting combination (either use BEGINOFF
	 * or initc) */
	if (initc[0] > 0) {
		for (i=0; i < TARGETN-2; i++)
			c[i] = initc[i];
	} else {
		for (i=0; i < TARGETN-2; i++)
			c[i] = i+BEGINOFF;
	}

	/* initialize all gmp integers we use */
	for (i=0; i < TARGETN-2+1; i++)
		mpz_init (&(v[i]));
	for (i = 0; i < ROWS1; i++) {
		for (j = 0; j < COLS1; j++) {
			mpz_init (&(gmp_submatrix[i][j]));
		}
	}

	/* precompute the modulo stuff */
	precompute_mod_tblP ();
	precompute_mod_matrixP (modP_matrix, the_matrix);

	/* print what combination we are starting at */
	printf("starting at: ");
	for (i = 0; i < TARGETN-2; i++) {
		printf ("%d", tr[c[i]-1]);
		if (i < TARGETN-2-1)
			printf (",");
	}
	printf("   (");
	for (i = 0; i < TARGETN-2; i++) {
		printf ("%d", c[i]);
		if (i < TARGETN-2-1)
			printf (",");
	}
	printf(")\n");

	fflush(stdout);

	/* Assuming x^d+y^d */
	do {
#ifdef ENDOFF
		if (__builtin_expect(c[0] > ENDOFF, 0))
			break;
#endif
		allcnt ++;

		/* must include one term of degree d-1 */
		if (tr[c[TARGETN-2-1]-1] < startof_dminusone) {
			notopcnt++;
			continue;
		}

		/* Must not include no adjecent terms
		 * such polynomials are never optimal */
		if (checkadjecent (c, TARGETN-2)) {
			adjcnt++;
			continue;
		}

		/* Get the mod MODULO submatrix and skip the iteration
		 * in the presence of null rows (null except for last column)
		 */
		if ( ! get_the_modP_submatrix (modP_submatrix, modP_matrix, the_matrix, c, TARGETN-2)) {
			/* all entries must affect the top degree terms */
			nrowcnt ++;
			goto status_update;
		}

		/* The submatrix must be singular, if not singular mod P
		 * then it won't be singular over rationals either */
		if (__builtin_expect(
		     modP_ref_is_full_rank (modP_submatrix),
		     TRUE)) {
			modfullrankcnt++;
			goto status_update;
		}

		/* get the matrix for the full elimination */
		gmp_get_the_submatrix (gmp_submatrix, the_matrix, c, TARGETN-2);

		finalcnt++;

		if (__builtin_expect(
			gmp_int_rref (gmp_submatrix, &p, TRUE /* dim_one_only */),
			FALSE)) {

			/* Null space is exactly one dimensional so get a
			 * single nullspace vector */
			gmp_null_vector (gmp_submatrix, v, &p);

			/* it must be all positive or all negatives */
			if (__builtin_expect(
			     (gmp_vector_nonnegative (v, TARGETN-2+1) ||
			      gmp_vector_nonpositive (v, TARGETN-2+1)),
			     FALSE)) {

				/* found an example, so rejoice and
				 * print it out! */

				printf ("******** EXAMPLE: ");
				for (i = 0; i < TARGETN-2+1; i++) {
					if (mpz_fits_slong_p (&(v[i]))) {
						long vi = mpz_get_si (&(v[i]));
						printf ("%ld", vi);
					} else {
						/* this should never happen */
						printf ("(too large)");
					}
					if (i < TARGETN-2)
						printf (",");
				}
				printf("   at: ");
				for (i = 0; i < TARGETN-2; i++) {
					printf ("%d", tr[c[i]-1]);
					if (i < TARGETN-2-1)
						printf (",");
				}
				printf("   (");
				for (i = 0; i < TARGETN-2; i++) {
					printf ("%d", c[i]);
					if (i < TARGETN-2-1)
						printf (",");
				}
				printf(")\n");
				fflush(stdout);
				cnt ++;
			}
		}

status_update:
		/* print updates in reasonable intervals */
		if (__builtin_expect(j == 10000000, 0)) {
			printf("at: ");
			for (i = 0; i < TARGETN-2; i++) {
				printf ("%d", tr[c[i]-1]);
				if (i < TARGETN-2-1)
					printf (",");
			}
			printf(" (");
			for (i = 0; i < TARGETN-2; i++) {
				printf ("%d", c[i]);
				if (i < TARGETN-2-1)
					printf (",");
			}
			printf(")\n");
			fflush(stdout);
			j=0;
		}
		j++;
	} while (__builtin_expect
		 (comb_get_next_combination (c, TARGETN-2, LOGCOLS),
		  TRUE));

	printf ("FOUND EXAMPLES: %d\n", cnt);
	fflush(stdout);

	printf ("allcnt: %Ld notopcnt: %Ld adjcnt: %Ld nrowcnt: %Ld "
		"modfullrankcnt: %Ld "
		"finalcnt: %Ld\n",
		allcnt, notopcnt, adjcnt, nrowcnt,
		modfullrankcnt,
		finalcnt);

	end_time = (int)time (NULL);
	printf ("ENDTIME: %d ELAPSED: %ds\n", end_time, end_time - begin_time);

	return 0;
}
